package io.github.iainvm.go.game;

import java.util.ArrayList;
import java.util.Arrays;

public class GameLogic {
    //This class contains all the logic for the game of Go "https://en.wikipedia.org/Go_(game)"
    //This class acts upon the given GameStates, meaning it doesn't need to be an Object Instance itself saving on
    //memory usage and reduces the chances of memory leaks


    //Get legal moves
    public static ArrayList<Integer> getLegalMoves(GameState gs){
        ArrayList<Integer> legalMoves = new ArrayList<>();

        for (int index = 0; index < gs.board.length; index++) {
            if((gs.board[index] == 0) && (isLegal(gs, index))){
                legalMoves.add(index);
            }
        }

        legalMoves.add(-1);

        return legalMoves;
    }

    //Check moves legality
    public static boolean isLegal(GameState gs, int move){
        if(move == -1){
            return true;
        }
        if(gs.board[move] != 0){
            return false;
        }

        boolean legal = false;

        //For each of the neighbouring counters
        for(int n : getNeighbours(gs, move)){
            //If it's got a free neighbour
            if(gs.board[n] == 0){
                legal = true;
            }
            //If it's the current players own counter
            if(!isOpponent(gs,n, gs.currentPlayer)){
                if(Math.abs(gs.board[n]) > 1){
                    //If it can attach to a chain that has a spare liberty
                    legal = true;
                }
            }else{
                //If it's the opponents counter
                if(Math.abs(gs.board[n]) == 1){
                    //Opponent's chain will disappear if the counter is placed, freeing up a liberty
                    legal = true;
                }
            }
        }

        //TODO: Check it's not the same state as the game 2 states back

        return legal;
    }

    //Make the given move
    public static GameState makeMove(GameState gamestate, int move){
        //Assume move is legal

        //Place counter in position
        GameState gs = new GameState(gamestate);

        gs.lastMove = move;

        if(move == -1){
            gs.passCounter++;
            gs.currentPlayer *= -1;
            return gs;
        }
        gs.passCounter = 0;
        gs.board[move] = gs.currentPlayer;

        if(GameState.debug){
            System.out.println("Neighbours: " + Arrays.toString(getNeighbours(gs, move).toArray()));
        }

        //For each Neighbour and self, do chain Check
        for(int n : getNeighbours(gs, move)){
            if(gs.board[n] != 0){
                chainCheck(gs, n);
            }
        }

        chainCheck(gs, move);

        //Switch current player
        gs.currentPlayer *= -1;

        return gs;
    }

    //Chain check
    private static void chainCheck(GameState gs, int index){
        int player = (gs.board[index] > 0)? 1 : -1;

        if(GameState.debug){System.out.println("Checking counter at position: " + index);}

        //Get Border
        ArrayList<Integer> border = getBorder(gs, index);

        if(GameState.debug){
            System.out.println("Border: " + Arrays.toString(border.toArray()));
        }


        //Count <Opp, Blank>
        int[] count = new int[2];
        for(int i : border){
            if(gs.board[i] == 0){
                count[1] ++;
            }else{
                if(isOpponent(gs, i, player)){
                    count[0] ++;
                }
            }
        }

        //Set value =  Blank
        setChain(gs, index, count[1]);

        //if value == 0 Chain Check border
        if(count[1] == 0){
            for(int i : border){
                chainCheck(gs, i);
            }
        }
    }

    //Checks if it's the opponents counter
    private static boolean isOpponent(GameState gs, int index, int player) {
        return !((player == 0) && (gs.board[index] == 0)) && !((player > 0) && (gs.board[index] > 0)) && !((player < 0) && (gs.board[index] < 0));
    }

    //Get the neighbours of a given point
    private static ArrayList<Integer> getNeighbours(GameState gs, int index){
        //Make neighbours array
        ArrayList<Integer> neighbours = new ArrayList<>();

        int val;

        //Check up (if you move up and aren't off the board)
        if((val = index - GameState.size) >= 0){
            neighbours.add(val);
        }
        //Check down (if you move down and aren't off the board)
        if((val = index + GameState.size) < gs.board.length){
            neighbours.add(val);
        }
        //Check left (if you move left and haven't wrapped round to the other side)
        if((val = index - 1)%GameState.size != GameState.size -1){
            if(val >= 0){
                neighbours.add(val);
            }
        }
        //Check right
        if((val = index + 1)%GameState.size != 0){
            neighbours.add(val);
        }

        return neighbours;
    }

    //Set chains value
    private static void setChain(GameState gs, int index, int val){
        int player = (gs.board[index] > 0)? 1 : -1;

        val *= player;

        for(int i: getChain(gs, index)){
            gs.board[i] = val;
        }
    }

    //Get all counters in chain (includes chains of empty points)
    private static ArrayList<Integer> getChain(GameState gs, int index){
        int player = (gs.board[index] > 0)? 1 : ((gs.board[index] < 0)? -1 : 0);


        //Set up chain array
        ArrayList<Integer> chain = new ArrayList<>();
        ArrayList<Integer> visited = new ArrayList<>();

        //Start the chain off
        chain.add(index);

        int i;
        int j = 0;
        //For every counter in the chain
        while(j < chain.size()){
            i = chain.get(j);

            //if it hasn't already been visited
            if(!visited.contains(i)){
                //check it's neighbours
                for(int n : getNeighbours(gs, i)){

                    //If it's part of the same chain AND not added already
                    if((!isOpponent(gs,n, player)) && (!chain.contains(n))){
                        chain.add(n);
                    }
                }
                visited.add(i);
            }

            j++;
        }

        if(GameState.debug){
            System.out.println("Chain: " + Arrays.toString(chain.toArray()));
        }
        return chain;
    }

    //Get neighbours of a chain
    private static ArrayList<Integer> getBorder(GameState gs, int index){
        ArrayList<Integer> border = new ArrayList<>();

        //Get the chain the given counter is in
        ArrayList<Integer> chain = getChain(gs, index);

        //For every counter in the chain
        for(int i : chain){
            //Grab it's neighbours
            for(int n : getNeighbours(gs, i)){
                //If that neighbour isn't in the chain AND hasn't been added to the border add it
                if((!chain.contains(n)) && (!border.contains(n))){
                    border.add(n);
                }
            }
        }
        if(GameState.debug){
            System.out.println("Border: " + Arrays.toString(border.toArray()));
        }
        return border;
    }

    //Check if game has ended
    public static boolean isFinished(GameState gs){
        return ((gs.passCounter == 2)||(getLegalMoves(gs).size() == 0));
    }

    //Score game
    public static int[] score(GameState gs){
        //score
        int[] score = {0,0};
        //Visited Array
        ArrayList<Integer> visited = new ArrayList<>();
        ArrayList<Integer> chain;

        ArrayList<Integer> border;
        //For every empty point on the board
        for(int i = 0; i < gs.board.length; i++){
            //If it's a blank area and not already visited
            if(gs.board[i] == 0 && !visited.contains(i)) {
                int[] borderCheck = new int[2];
                border = getBorder(gs, i);

                //Tally up what the border counters are
                for (int b : border) {
                    if (gs.board[b] > 0) {
                        //Black counter
                        borderCheck[0]++;
                    }
                    if (gs.board[b] < 0) {
                        //White counter
                        borderCheck[1]++;
                    }
                }

                chain = getChain(gs, i);
                visited.addAll(chain);

                if(borderCheck[0] == borderCheck[1]) {
                    if (borderCheck[1] == 0) {
                        //Black wins area
                        score[0] = chain.size();
                    }
                    if (borderCheck[0] == 0) {
                        score[1] = chain.size();
                    }
                }
            }
            if(gs.board[i] > 0){
                score[0] ++;
            }
            if(gs.board[i] < 0){
                score[1] ++;
            }
        }
        return score;
    }
}