package io.github.iainvm.go.players.ai;

import io.github.iainvm.go.game.GameLogic;
import io.github.iainvm.go.game.GameState;
import io.github.iainvm.go.players.Player;

import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Map;
import java.util.Random;

public abstract class AI implements Runnable, Player {

    private TreeNode root, current;
    private int pID, counter;
    private boolean threadRunning;

    //Constructor
    public AI(int pID){
        //Set pID
        this.pID = pID;
    }

    //Decide a move from a given state
    public int getMove(GameState gs){
        //Add the given gs to the tree if it's not already there

        //If the gamestate isn't the root (already added) and doesn't already exist as a child
        if(root == null){
            root = new TreeNode(null, new GameState(gs));
            current = root;
        }else{
            current.children.put(gs.lastMove, new TreeNode(current, new GameState(gs)));
            //Make new node current node
            current = current.children.get(gs.lastMove);
        }


        //Start thread to simulate those moves, and set <win,plays>
        Thread t = new Thread(this);
        t.start();

        int min = 0;
        int sec = 10;
        //Kill thread after given amount of time (SAFELY)
        long startTime = System.currentTimeMillis()/1000L;
        long nowTime = System.currentTimeMillis()/1000L;

        while(nowTime - startTime < ((min*60)+sec)){
            nowTime = System.currentTimeMillis()/1000L;
        }
        killThread();

        //Select best move
        return selectMove();
    }

    //Picks the best move from the children of the current Node
    private int selectMove(){
        return selectBest(current).gs.lastMove;
    }

    //Picks the UCB node for iterating down the tree
    private TreeNode selectUCB(TreeNode tn){
        Random rnd = new Random();
        double bestScore = 0;
        double score;
        double C = Math.sqrt(2);
        ArrayList<TreeNode> nodes = new ArrayList<>();

        for(Map.Entry<Integer, TreeNode> child : tn.children.entrySet()){
            score = (child.getValue().wins)/(child.getValue().plays) + (C*Math.sqrt(Math.log((tn.plays/child.getValue().plays))));

            if(score > bestScore){
                bestScore = score;
                nodes = new ArrayList<>();
                nodes.add(child.getValue());
            }
            if(score == bestScore){
                nodes.add(child.getValue());
            }
        }

        return nodes.get(rnd.nextInt(nodes.size()));
    }

    //Picks the UCB node for iterating down the tree
    private TreeNode selectOppUCB(TreeNode tn){
        Random rnd = new Random();
        double bestScore = 99999;
        double score;
        double C = Math.sqrt(2);
        ArrayList<TreeNode> nodes = new ArrayList<>();



        for(Map.Entry<Integer, TreeNode> child : tn.children.entrySet()){
            score = (child.getValue().wins)/(child.getValue().plays) + (C*Math.sqrt(Math.log((tn.plays/child.getValue().plays))));

            if(score < bestScore){
                bestScore = score;
                nodes = new ArrayList<>();
                nodes.add(child.getValue());
            }
            if(score == bestScore){
                nodes.add(child.getValue());
            }
        }

        return nodes.get(rnd.nextInt(nodes.size()));
    }

    //Picks the best move from the children of the current Node
    private TreeNode selectBest(TreeNode tn){
        Random rnd = new Random();
        double bestScore = 0;
        double score;
        ArrayList<TreeNode> nodes = new ArrayList<>();

        for(Map.Entry<Integer, TreeNode> child : tn.children.entrySet()){
            score = (child.getValue().wins)/(child.getValue().plays);

            if(score > bestScore){
                bestScore = score;
                nodes = new ArrayList<>();
                nodes.add(child.getValue());
            }
            if(score == bestScore){
                nodes.add(child.getValue());
            }
        }

        return nodes.get(rnd.nextInt(nodes.size()));
    }

    //Create many children of current node to simulate
    public abstract void expandChildren(TreeNode tn);

    //Is it a leaf
    private boolean isLeaf(TreeNode tn){
        return tn.children.size() == 0;
    }

    //Is it a terminal node
    private boolean isTerminal(TreeNode tn){
        //TODO: Improve
        //Problem: Game was randomly selecting 2 passes, so causing
        return GameLogic.isFinished(tn.gs);
//        return GameLogic.getLegalMoves(tn.gs).size() == 0;
    }

    //Select optimal leaf
    private TreeNode optimalLeaf(TreeNode tn){
        TreeNode pointer = tn;

        while(!isLeaf(pointer)){
            pointer = selectUCB(pointer);
            if(!isLeaf(pointer)){
                pointer = selectOppUCB(pointer);
            }
        }

        return pointer;
    }

    //Simulate from the given node
    private void simulate(TreeNode pointer){
        Random rnd = new Random();
        GameState gs = new GameState(pointer.gs);
        ArrayList<Integer> legalMoves;
        int move;

        //While it's not a terminal node
        while((!GameLogic.isFinished(gs))&&(threadRunning)){
            //Pick a random child
            legalMoves = GameLogic.getLegalMoves(gs);
            move = legalMoves.get(rnd.nextInt(legalMoves.size()));

            gs = GameLogic.makeMove(gs, move);
        }

        //Score
        int[] score = GameLogic.score(gs);
        int playerIndex = (current.gs.currentPlayer > 0)? 0 : 1;
        boolean won = score[playerIndex] > score[(playerIndex+1)%2];
        pointer.plays++;
        if(won){
            pointer.wins++;
        }

        //Iterate back up the tree
        while(pointer.parent != null){
            pointer = pointer.parent;
            pointer.plays++;
            if(won){
                pointer.wins++;
            }
        }
    }

    //Expands the tree from the current TreeNode
    @Override
    public void run() {
        threadRunning = true;
        TreeNode pointer;
        counter = 0;

        while(threadRunning){
            //From current node
            //Select optimal child nodes down to leaf node
            counter++;

            pointer = optimalLeaf(current);
            //If L isn't terminal (isFinished) create child
            if(!isTerminal(pointer)){
                expandChildren(pointer);
            }

            //Simulate child node
            for(Map.Entry<Integer, TreeNode> child : pointer.children.entrySet()){
                simulate(child.getValue());
            }
        }
    }

    //Kill the running thread
    private void killThread(){
        System.out.println("Player " + pID + " made " + counter + " Loops");
        try {
            System.out.println("Writer Check");
            FileWriter fw = new FileWriter("C:\\Users\\IainV\\IdeaProjects\\GoGame\\out\\production\\GoGame\\speed" + pID + ".csv", true);

            fw.append(GameLogic.getLegalMoves(current.gs).size() + "," + counter + "\n");
        } catch (IOException e) {
            e.printStackTrace();
        }
        System.out.println("Current probability " + current.wins/current.plays);
        this.threadRunning = false;
    }
}
